@@ -283,7 +283,7 @@ def magic_token(length=token_length):
     return random.choice(string.ascii_lowercase) + functools.reduce(lambda x, y: x + random.choice(token_chars), range(length-1), '').lower()
 
 def validate_token(row_id, token):
-    return validate_int(row_id) and len([ch for ch in token if ch in token_chars]) == token_length
+    return validate_int(row_id) and token and len([ch for ch in token if ch in token_chars]) == token_length
 
 def validate_int(value):
     # int() can't take a None so we have to check that first
@@ -332,6 +332,14 @@ def validate_ip_address(value):
     except:
         return None
 
+def validate_json(self, value):
+    if type(value) is dict:
+        return value
+    try:
+        return json.loads(value)
+    except ValueError:
+        return None
+
 
 email_validator = None
 def validate_email(value):
@@ -836,3 +844,4 @@ def incr_stop(obj, timer_name):
     obj._timers.setdefault(cnt_name, 0)
     obj._timers[exec_name] += obj._timers[stop_name] - obj._timers[start_name]
     obj._timers[cnt_name] += 1
+    return obj._timers[stop_name] - obj._timers[start_name]
